"""Radiotap"""
import logging

from packetracer import packetracer, triggerlist
# handler
from packetracer.layer12 import ieee80211
from packetracer.structcbs import pack_H_le, unpack_H_le, unpack_I

logger = logging.getLogger("packetracer")

RTAP_TYPE_80211 = 0

# Ref: http://www.radiotap.org
# Fields Ref: http://www.radiotap.org/defined-fields/all

# defined flags ordered by appearance (big endian)
TSFT_MASK = 0x01000000
FLAGS_MASK = 0x02000000
RATE_MASK = 0x04000000
CHANNEL_MASK = 0x08000000

FHSS_MASK = 0x10000000
DB_ANT_SIG_MASK = 0x20000000
DB_ANT_NOISE_MASK = 0x40000000
LOCK_QUAL_MASK = 0x80000000

TX_ATTN_MASK = 0x00010000
DB_TX_ATTN_MASK = 0x00020000
DBM_TX_POWER_MASK = 0x00040000
ANTENNA_MASK = 0x00080000

ANT_SIG_MASK = 0x00100000
ANT_NOISE_MASK = 0x00200000
RX_FLAGS_MASK = 0x00400000

CHANNELPLUS_MASK = 0x00000400
HT_MASK = 0x00000800

AMPDU_MASK = 0x00001000
VHT_MASK = 0x00002000

# 7 bits reserved

RT_NS_NEXT_MASK = 0x00000020
VENDOR_NS_NEXT = 0x00000040
EXT_MASK = 0x00000080

# mask -> (length, alignment)
RADIO_FIELDS = {
    TSFT_MASK: (8, 8),
    FLAGS_MASK: (1, 1),
    RATE_MASK: (1, 1),
    # channel + flags
    CHANNEL_MASK: (4, 2),

    # fhss + pattern
    FHSS_MASK: (2, 1),
    DB_ANT_SIG_MASK: (1, 1),
    DB_ANT_NOISE_MASK: (1, 1),
    LOCK_QUAL_MASK: (2, 2),

    TX_ATTN_MASK: (2, 2),
    DB_TX_ATTN_MASK: (2, 2),
    DBM_TX_POWER_MASK: (1, 1),
    ANTENNA_MASK: (1, 1),

    ANT_SIG_MASK: (1, 1),
    ANT_NOISE_MASK: (1, 1),
    RX_FLAGS_MASK: (2, 2),

    # CHANNELPLUS_MASK	:,
    HT_MASK: (3, 1),

    AMPDU_MASK: (8, 4),
    VHT_MASK: (12, 2)

    # RT_NS_NEXT_MASK	:,
    # VENDOR_NS_NEXT	:,
    # EXT_MASK		:
}

RADIO_FIELDS_MASKS = [
    TSFT_MASK,
    FLAGS_MASK,
    RATE_MASK,
    # channel + flags
    CHANNEL_MASK,

    # fhss + pattern
    FHSS_MASK,
    DB_ANT_SIG_MASK,
    DB_ANT_NOISE_MASK,
    LOCK_QUAL_MASK,

    TX_ATTN_MASK,
    DB_TX_ATTN_MASK,
    DBM_TX_POWER_MASK,
    ANTENNA_MASK,

    ANT_SIG_MASK,
    ANT_NOISE_MASK,
    RX_FLAGS_MASK,

    HT_MASK,

    AMPDU_MASK,
    VHT_MASK
]


class FlagTriggerList(triggerlist.TriggerList):
    # no __init__ needed: we just add tuples
    def _pack(self, tuple_entry):
        # return b"".join([flag[1] for flag in self])
        return tuple_entry[1]


def get_channelinfo(channel_bytes):
    """
    return -- [channel_mhz, channel_flags]
    """
    return [unpack_H_le(channel_bytes[0:2])[0], unpack_H_le(channel_bytes[2:4])[0]]


def freq_to_channel(freq):
    """
    freq -- frequqncy in Hz
    return -- channel number
    """
    if 2412000000 <= freq <= 2472000000:
        return 1 + int((freq - 2412000000) / (5 * 1000000))
    if freq == 2484000000:
        return 14
    if 5035000000 <= freq <= 5825000000:
        return 7 + int((freq - 5035000000) / (5 * 1000000))
    return None


def channel_to_freq(channel):
    """
    freq -- frequqncy in Hz
    return -- channel number
    """
    if 1 <= channel <= 13:
        return 2407000000 + channel * 5 * 1000000
    if channel == 14:
        return 2484000000
    if channel >= 15:
        return 5035000000 + (channel - 7) * 5 * 1000000
    return None


class Radiotap(packetracer.Packet):
    __hdr__ = (
        ("version", "B", 0),
        ("pad", "B", 0),
        ("len", "H", 0x0800),
        ("present_flags", "I", 0),
        ("flags", None, FlagTriggerList)  # stores: (XXX_MASK, value)
    )

    __handler__ = {
        RTAP_TYPE_80211: ieee80211.IEEE80211
    }

    # handle frame check sequence
    def _get_fcs(self):
        try:
            return self._fcs
        except AttributeError:
            return b""

    def _set_fcs(self, fcs):
        self._fcs = fcs

    fcs = property(_get_fcs, _set_fcs)

    def _get_channel(self):
        return self.flags.get_by_key(CHANNEL_MASK)

    def _set_channel(self, channel):
        freq = int(channel_to_freq(channel) / 1000000)
        self.flags.set_by_key(CHANNEL_MASK, pack_H_le(freq))

    # get/set channel, set frequency under the hood
    channel = property(_get_channel, _set_channel)

    def _dissect(self, buf):
        # Needed by _parse_flags()
        self.present_flags = unpack_I(buf[4:8])[0]
        # logger.debug("Flags: 0x%X" % flags)
        pos_end = len(buf)

        # Check for FSC, needs to be skipped for upper layers
        if self.present_flags & FLAGS_MASK == FLAGS_MASK:
            # logger.debug("Flags mask matched")
            flags_off = 8
            # TSFT present -> Flags values are after TSFT value
            if self.present_flags & TSFT_MASK == TSFT_MASK:
                flags_off = 8 + 8
            if buf[flags_off] & 0x10 != 0:
                # logger.debug("FCS found")
                self._fcs = buf[-4:]
                pos_end = -4

        hdr_len = unpack_H_le(buf[2:4])[0]
        # logger.debug("Bytes for flags (%d): %s" % (hdr_len, buf[8: hdr_len]))
        self._init_triggerlist("flags", buf[8: hdr_len], self._parse_flags)
        # logger.debug("rtap bytes:=%r" % buf[:hdr_len])
        # now we got the correct header length
        self._init_handler(RTAP_TYPE_80211, buf[hdr_len: pos_end])
        # logger.debug("Adding %d flags" % len(self.flags))
        return hdr_len

    def _parse_flags(self, buf):
        off = 0
        flags = []

        # Assume order of flags is correctly stated by "present_flags"
        # We need to know if fcs is present: minimum TSFT and flags must get parsed
        for mask in RADIO_FIELDS_MASKS:
            # logger.debug(self.present_flags)
            # flag not set
            if mask & self.present_flags == 0:
                continue

            size_align = RADIO_FIELDS[mask]
            size = size_align[0]
            # check alignment
            mod = off % size_align[1]

            if mod != 0:
                # enlarge size by alignment
                size += (size_align[1] - mod)

            # logger.debug("got flag %02X, length/align: %r" % (mask, size_align))
            # add all fields for the stated flag
            value = buf[off: off + size]

            # FCS present?
            # if mask == FLAGS_MASK and unpack_B(value)[0] & 0x10 != 0:
            #	# logger.debug("fcs found")
            #	fcs_present = True

            # logger.debug("adding flag: %s" % str(mask))
            flags.append((mask, value))
            off += size
        return flags

    def bin(self, update_auto_fields=True):
        """Custom bin(): handle FCS."""
        return packetracer.Packet.bin(self, update_auto_fields=update_auto_fields) + self.fcs



"""
Packet read and write routines for pcap format.
See http://wiki.wireshark.org/Development/LibpcapFileFormat
"""
import logging
import types

from packetracer import packetracer
from packetracer.layer12 import ethernet, linuxcc, radiotap, btle, can
from packetracer.structcbs import pack_I, unpack_I_le
from packetracer.structcbs import pack_IIII, unpack_IIII, unpack_IIII_le

logger = logging.getLogger("packetracer")

# PCAP/TCPDump related
# PCAP file header

# File magic numbers
# pcap using microseconds resolution
TCPDUMP_MAGIC = 0xA1B2C3D4
TCPDUMP_MAGIC_SWAPPED = 0xD4C3B2A1
# pcap using nanoseconds resolution
TCPDUMP_MAGIC_NANO = 0xA1B23C4D
TCPDUMP_MAGIC_NANO_SWAPPED = 0x4D3CB2A1

PCAP_VERSION_MAJOR = 2
PCAP_VERSION_MINOR = 4

DLT_NULL = 0
DLT_EN10MB = 1
DLT_EN3MB = 2
DLT_AX25 = 3
DLT_PRONET = 4
DLT_CHAOS = 5
DLT_IEEE802 = 6
DLT_ARCNET = 7
DLT_SLIP = 8
DLT_PPP = 9
DLT_FDDI = 10
DLT_PFSYNC = 18
DLT_IEEE802_11 = 105
DLT_LINUX_SLL = 113
DLT_PFLOG = 117
DLT_IEEE802_11_RADIO = 127
DLT_CAN_SOCKETCAN = 227
DLT_LINKTYPE_BLUETOOTH_LE_LL = 251
LINKTYPE_BLUETOOTH_LE_LL_WITH_PHDR = 256

PCAPTYPE_CLASS = {
    DLT_LINUX_SLL: linuxcc.LinuxCC,
    DLT_EN10MB: ethernet.Ethernet,
    DLT_CAN_SOCKETCAN: can.CAN,
    DLT_IEEE802_11_RADIO: radiotap.Radiotap,
    LINKTYPE_BLUETOOTH_LE_LL_WITH_PHDR: btle.BTLEHdr
}


class PcapFileHdr(packetracer.Packet):
    """pcap file header."""
    # header length = 24
    __hdr__ = (
        ("magic", "I", TCPDUMP_MAGIC),
        ("v_major", "H", PCAP_VERSION_MAJOR),
        ("v_minor", "H", PCAP_VERSION_MINOR),
        ("thiszone", "I", 0),
        ("sigfigs", "I", 0),
        ("snaplen", "I", 1500),
        ("linktype", "I", 1),
    )


class PcapPktHdr(packetracer.Packet):
    """pcap packet header."""
    # header length: 16
    __hdr__ = (
        ("tv_sec", "I", 0),
        # this can be either microseconds or nanoseconds: check magic number
        ("tv_usec", "I", 0),
        ("caplen", "I", 0),
        ("len", "I", 0),
    )


# PCAP callbacks


def pcap_cb_init_write(self, snaplen=1500, linktype=DLT_EN10MB, **initdata):
    self._timestamp = 0
    header = PcapFileHdr(magic=TCPDUMP_MAGIC_NANO, snaplen=snaplen, linktype=linktype)
    self._fh.write(header.bin())


def pcap_cb_write(self, bts, **metadata):
    ts = metadata.get("ts", self._timestamp + 1000000)
    self._timestamp = ts
    sec = int(ts / 1000000000)
    nsec = ts - (sec * 1000000000)

    # logger.debug("paket time sec/nsec: %d/%d", sec, nsec)
    n = len(bts)
    self._fh.write(pack_IIII(sec, nsec, n, n))
    self._fh.write(bts)


def pcap_cb_init_read(self, **initdata):
    buf = self._fh.read(24)
    # file header is skipped per default (needed for __next__)
    self._fh.seek(24)
    # this is not needed anymore later on but we set it anyway
    fhdr = PcapFileHdr(buf)
    self._closed = False

    if fhdr.magic not in [TCPDUMP_MAGIC, TCPDUMP_MAGIC_NANO, TCPDUMP_MAGIC_SWAPPED, TCPDUMP_MAGIC_NANO_SWAPPED]:
        return False
    is_le = False

    # handle file types
    if fhdr.magic == TCPDUMP_MAGIC:
        self._resolution_factor = 1000
        # Note: we could use PcapPktHdr/PcapLEPktHdr to parse pre-packetdata but calling unpack directly
        # greatly improves performance
        self._callback_unpack_meta = unpack_IIII
    elif fhdr.magic == TCPDUMP_MAGIC_NANO:
        self._resolution_factor = 1
        self._callback_unpack_meta = unpack_IIII
    elif fhdr.magic == TCPDUMP_MAGIC_SWAPPED:
        is_le = True
        self._resolution_factor = 1000
        self._callback_unpack_meta = unpack_IIII_le
    elif fhdr.magic == TCPDUMP_MAGIC_NANO_SWAPPED:
        is_le = True
        self._resolution_factor = 1
        self._callback_unpack_meta = unpack_IIII_le
    else:
        raise ValueError("invalid tcpdump header, magic value: %s" % fhdr.magic)

    linktype = fhdr.linktype if not is_le else unpack_I_le(pack_I(fhdr.linktype))[0]
    self._lowest_layer_new = PCAPTYPE_CLASS.get(linktype, None)

    def is_resolution_nano(obj):
        """return -- True if resolution is in Nanoseconds, False if milliseconds."""
        return obj._resolution_factor == 1000

    self.is_resolution_nano = types.MethodType(is_resolution_nano, self)
    return True


def pcap_cb_read(self):
    buf = self._fh.read(16)

    if not buf:
        raise StopIteration

    d = self._callback_unpack_meta(buf)
    buf = self._fh.read(d[2])

    return d[0] * 1000000000 + (d[1] * self._resolution_factor), buf


def pcap_cb_btstopkt(self, meta, bts):
    return self._lowest_layer_new(bts)


# PCAPNG related_btstopkt
# Generic/filetype invariant related

FILETYPE_PCAP = 0
FILETYPE_PCAPNG = 1

# type_id : [
#	cb_init_write(obj, **initdata),
#	cb_write(self, bytes, **metadata),
#	cb_init_read(obj, **initdata),
#	cb_read(self): metadata, bytes
#	cb_btstopkt(self, metadata, bytes): pkt
# ]
FILEHANDLER = {
    FILETYPE_PCAP: [
        pcap_cb_init_write, pcap_cb_write, pcap_cb_init_read, pcap_cb_read, pcap_cb_btstopkt
    ]
    # FILETYPE_PCAPNG : [
    #	None, None, None
    # ]
}


class FileHandler(object):
    def __init__(self, filename, accessmode):
        self._fh = open(filename, accessmode)
        self._closed = False

    def __enter__(self):
        return self

    def __exit__(self, objtype, value, traceback):
        self.close()

    def flush(self):
        self._fh.flush()

    def close(self):
        self._closed = True
        self._fh.close()


class PcapHandler(FileHandler):
    MODE_READ = 1
    MODE_WRITE = 2

    def __init__(self, filename, mode, filetype=FILETYPE_PCAP, **initdata):
        try:
            callbacks = FILEHANDLER[filetype]
        except IndexError:
            raise Exception("unknown filehandler type for mode %d: %d" % (mode, filetype))

        if mode == PcapHandler.MODE_WRITE:
            super().__init__(filename, "wb")
            callbacks[0](self, **initdata)
            self.write = types.MethodType(callbacks[1], self)
        elif mode == PcapHandler.MODE_READ:
            super().__init__(filename, "rb")
            ismatch = False

            for pcaptype, callbacks in FILEHANDLER.items():
                self._fh.seek(0)
                # init callback
                ismatch = callbacks[2](self, **initdata)

                if ismatch:
                    # logger.debug("found handler for file: %x", pcaptype)
                    # read callback
                    self.__next__ = types.MethodType(callbacks[3], self)
                    # bytes-to-packet callback
                    self._btstopkt = types.MethodType(callbacks[4], self)
                    break
            if not ismatch:
                raise Exception("no matching handler found")
        else:
            raise Exception("wrong mode: %d" % mode)

    def read_packet(self, pktfilter=lambda pkt: True):
        """
        pktfilter -- filter as lambda function to match packets to be retrieved,
            return True to accept a specific packet.
        return -- (metadata, packet) if packet can be created from bytes
            else (metadata, bytes). For pcap/tcpdump metadata is a nanoseconds timestamp
        """
        while True:
            # until StopIteration
            meta, bts = self.__next__()

            try:
                pkt = self._btstopkt(meta, bts)
            except Exception as ex:
                logger.warning("could not create packets from bytes: %r", ex)
                return meta, bts

            if pktfilter(pkt):
                return meta, pkt

    def read_packet_iter(self, pktfilter=lambda pkt: True):
        """
        pktfilter -- filter as lambda function to match packets to be retrieved,
            return True to accept a specific packet.
        return -- iterator yielding (metadata, packet)
        """
        if self._closed:
            return

        while True:
            try:
                yield self.read_packet(pktfilter=pktfilter)
            except:
                return

    def __iter__(self):
        """
        return -- (metadata, bytes)
        """
        if self._closed:
            return

        while True:
            try:
                yield self.__next__()
            except StopIteration:
                break


class Writer(PcapHandler):
    """
    Simple pcap writer supporting pcap format.
    """

    def __init__(self, filename, filetype=FILETYPE_PCAP, **initdata):
        super().__init__(filename, PcapHandler.MODE_WRITE, **initdata)


class Reader(PcapHandler):
    """
    Simple pcap file reader supporting pcap format.
    """

    def __init__(self, filename, filetype=FILETYPE_PCAP, **initdata):
        super().__init__(filename, PcapHandler.MODE_READ, **initdata)



"""
Message Queuing Telemetry Transport (MQTT)
https://docs.oasis-open.org/mqtt/mqtt/v3.1.1/os/mqtt-v3.1.1-os.html
"""
import logging

from packetracer.packetracer import Packet
from packetracer.structcbs import pack_B, unpack_H

logger = logging.getLogger("packetracer")

# Message Types:
MSGTYPE_RESERVED = 0
MSGTYPE_CONNECT = 1
MSGTYPE_CONNACK = 2
MSGTYPE_PUBLISH = 3
MSGTYPE_PUBACK = 4
MSGTYPE_PUBRECV = 5
MSGTYPE_PUBREL = 6
MSGTYPE_PUBCOMPLETE = 7
MSGTYPE_SUBSCRIBEREQ = 8
MSGTYPE_SUBSCRIBEACK = 9
MSGTYPE_UNSUBSCRIBE = 10
MSGTYPE_UNSUBACK = 11
MSGTYPE_PINGREQ = 12
MSGTYPE_PINGRESP = 13
MSGTYPE_DISCONNECT = 14


class Connect(Packet):
    __hdr__ = (
        ("pnamelen", "H", 0),
        ("pname", None, b""),
        ("version", "B", 0),
        ("conflags", "B", 0),
        ("keepalive", "H", 0),
        ("clientidlen", "H", 0),
        ("clientid", None, b"")
    )

    def _dissect(self, buf):
        pnamelen = unpack_H(buf[:2])[0]
        self.pname = buf[2: 2 + pnamelen]
        off_clientidlen = 2 + pnamelen + 1 + 1 + 2
        clientidlen = unpack_H(buf[off_clientidlen: off_clientidlen + 2])[0]
        off_clientid = off_clientidlen + 2
        self.clientid = buf[off_clientid: off_clientid + clientidlen]
        return 8 + pnamelen + clientidlen


class ConnAck(Packet):
    __hdr__ = (
        ("reserved", "B", 0),
        ("retcode", "B", 0)
    )


class Publish(Packet):
    __hdr__ = (
        ("topiclen", "H", 0),
        ("topic", None, b"")
    )

    def _dissect(self, buf):
        topiclen = unpack_H(buf[:2])[0]
        self.topic = buf[2: 2 + topiclen]

        return 2 + topiclen


class PubAck(Packet):
    __hdr__ = (
    )


class PubRecv(Packet):
    __hdr__ = (
        ("msgid", "H", 0),
    )


class PubRel(Packet):
    __hdr__ = (
        ("msgid", "H", 0),
    )


class PubComplete(Packet):
    __hdr__ = (
        ("msgid", "H", 0),
    )


class SubRequest(Packet):
    __hdr__ = (
        ("msgid", "H", 0),
        ("topiclen", "H", 0),
        ("topic", None, b""),
        ("qos", "B", 0)
    )

    def _dissect(self, buf):
        topiclen = unpack_H(buf[2: 4])[0]
        self.topic = buf[4: 4 + topiclen]

        return 5 + topiclen


class SubAck(Packet):
    __hdr__ = (
        ("msgid", "H", 0),
        ("qos", "B", 0)
    )


class Unsubscribe(Packet):
    __hdr__ = (
    )


class UnsubAck(Packet):
    __hdr__ = (
    )


class PingReq(Packet):
    __hdr__ = (
    )


class PingResp(Packet):
    __hdr__ = (
    )


class Discconnect(Packet):
    __hdr__ = (
    )


class MQTTBase(Packet):
    __hdr__ = (
        ("flags", "B", 1),
        ("mlen", None, b"\x00")  # 0xF000 = 11110000 00000000 = [one more byte] 1110000 [no more byte] 0000000
    )

    __handler__ = {
        MSGTYPE_CONNECT: Connect,
        MSGTYPE_CONNACK: ConnAck,
        MSGTYPE_PUBLISH: Publish,
        MSGTYPE_PUBACK: PubAck,
        MSGTYPE_PUBRECV: PubRecv,
        MSGTYPE_PUBREL: PubRel,
        MSGTYPE_PUBCOMPLETE: PubComplete,
        MSGTYPE_SUBSCRIBEREQ: SubRequest,
        MSGTYPE_SUBSCRIBEACK: SubAck,
        MSGTYPE_UNSUBSCRIBE: Unsubscribe,
        MSGTYPE_UNSUBACK: UnsubAck,
        MSGTYPE_PINGREQ: PingReq,
        MSGTYPE_PINGRESP: PingResp,
        MSGTYPE_DISCONNECT: Discconnect
    }

    def _dissect(self, buf):
        # Length MUST be decoded, flexible format but more imperformant bc parsing needed
        mlen_len, _ = MQTTBase._decode_length(buf[1:])
        self.mlen = buf[1: 1 + mlen_len]
        hlen = 1 + mlen_len
        # logger.debug("Init handler..., mlen: %s, id: %d, buf: %s" % (mlen_len, (buf[0] & 0xF0) >> 4, buf[hlen:]))

        try:
            self._init_handler((buf[0] & 0xF0) >> 4, buf[hlen:])
        except:
            # no type found
            pass
        return hlen

    @staticmethod
    def _decode_length(buf):
        """return -- mlen length, mlen value"""
        if buf[0] == 0x00:
            return 1, 0
        buf_idx = 0
        current_bt = buf[buf_idx]
        retval = current_bt & 0x7F

        while current_bt & 0x80 != 0 and buf_idx < len(buf):
            buf_idx += 1
            current_bt = buf[buf_idx]
            retval += (current_bt & 0x7F) << 7 * buf_idx
        return 1 + buf_idx, retval

    @staticmethod
    def _encode_length(num):
        """
        num -- Positive integer like 256
        return -- Encoded number as bytes like b"\x81\xff"
        """
        bts = []
        while num > 0:
            tbenc = num % 128
            num = int(num / 128)
            hbit = 0x80 if num > 0 else 0
            bts.append(pack_B(hbit | tbenc))

        return b"".join(bts)

    def _get_mlen(self):
        return MQTTBase._decode_length(self.mlen)

    def _set_mlen(self, val):
        self.mlen = MQTTBase._encode_length(val)

    mlen_d = property(_get_mlen, _set_mlen)

    def _get_msgtype(self):
        return (self.flags & 0xF0) >> 4

    def _set_msgtype(self, val):
        self.flags = (val & 0x0F) << 4 | (self.flags & 0x0F)

    msgtype = property(_get_msgtype, _set_msgtype)



"""
ATA over Ethernet
http://brantleycoilecompany.com/AoEr11.pdf
See https://en.wikipedia.org/wiki/ATA_over_Ethernet
"""

from packetracer import packetracer


class AOECFG(packetracer.Packet):
    __hdr__ = (
        ("bufcnt", "H", 0),
        ("fwver", "H", 0),
        ("scnt", "B", 0),
        ("aoeccmd", "B", 0),
        ("cslen", "H", 0),
    )


ATA_DEVICE_IDENTIFY = 0xEC


class AOEATA(packetracer.Packet):
    __hdr__ = (
        ("aflags", "B", 0),
        ("errfeat", "B", 0),
        ("scnt", "B", 0),
        ("cmdstat", "B", ATA_DEVICE_IDENTIFY),
        ("lba0", "B", 0),
        ("lba1", "B", 0),
        ("lba2", "B", 0),
        ("lba3", "B", 0),
        ("lba4", "B", 0),
        ("lba5", "B", 0),
        ("res", "H", 0)
    )


class AOE(packetracer.Packet):
    __hdr__ = (
        ("ver_fl", "B", 0x10),
        ("err", "B", 0),
        ("maj", "H", 0),
        ("min", "B", 0),
        ("cmd", "B", 0),
        ("tag", "I", 0)
    )

    def _get_ver(self):
        return self.ver_fl >> 4

    def _set_ver(self, ver):
        self.ver_fl = (ver << 4) | (self.ver_fl & 0xF)

    ver = property(_get_ver, _set_ver)

    def _get_fl(self):
        return self.ver_fl & 0xF

    def _set_fl(self, fl):
        self.ver_fl = (self.ver_fl & 0xF0) | fl

    fl = property(_get_fl, _set_fl)



"""Telnet."""
from packetracer import packetracer, triggerlist
from packetracer.structcbs import pack_B

IAC = 255  # interpret as command:
DONT = 254  # you are not to use option
DO = 253  # please, you use option
WONT = 252  # I won"t use option
WILL = 251  # I will use option
SB = 250  # interpret as subnegotiation
GA = 249  # you may reverse the line
EL = 248  # erase the current line
EC = 247  # erase the current character
AYT = 246  # are you there
AO = 245  # abort output--but let prog finish
IP = 244  # interrupt process--permanently
BREAK = 243  # break
DM = 242  # data mark--for connect. cleaning
NOP = 241  # nop
SE = 240  # end sub negotiation
EOR = 239  # end of record (transparent mode)
ABORT = 238  # Abort process
SUSP = 237  # Suspend process
xEOF = 236  # End of file: EOF is already used...

SYNCH = 242  # for telfunc calls

TELNET_OPTION_START = b"\xff\xaa"
TELNET_OPTION_END = b"\xff\x00"


class Telnet(packetracer.Packet):
    __hdr__ = (
        ("telnet_data", None, triggerlist.TriggerList),
    )

    def _dissect(self, buf):
        self._init_triggerlist("telnet_data", buf, self._parse_data)
        return len(buf)

    @staticmethod
    def _parse_data(buf):

        off = 0
        t_data = []
        t_len = len(buf)

        # parse telnet data:
        # fffaXX = start of options
        # fff0 = end of options
        while off < t_len:
            if buf[off: off + 2] == TELNET_OPTION_START:
                # add start marker
                t_data.append(buf[off: off + 3])
                off += 3
                # find end of option
                idx_end = buf.find(TELNET_OPTION_END, off)
                # add option data
                t_data.append(buf[off: idx_end + 1])
                # add end marker
                t_data.append(TELNET_OPTION_END)
                off = idx_end + 2
            else:
                # add command
                t_data.append(buf[off: off + 3])
                off += 3
        return t_data


def strip_options(buf):
    """Return a list of lines and dict of options from telnet data."""
    tokens = buf.split(pack_B(IAC))
    b = []
    d = {}
    subopt = False
    for w in tokens:
        if not w:
            continue
        o = w[0]
        if o > SB:
            # print("WILL/WONT/DO/DONT/IAC", "w")
            w = w[2:]
        elif o == SE:
            # print("SE", "w")
            w = w[1:]
            subopt = False
        elif o == SB:
            # print("SB", "w")
            subopt = True
            for opt in (b"USER", b"DISPLAY", b"TERM"):
                p = w.find(opt + b"\x01")
                if p != -1:
                    d[opt] = w[p + len(opt) + 1:].split(b"\x00", 1)[0]
            w = None
        elif subopt:
            w = None
        if w:
            w = w.replace(b"\x00", b"\n").splitlines()
            if not w[-1]:
                w.pop()
            b.extend(w)
    return b, d
